import matlab.engine
import itertools
import numpy as np
import pytest

from pynrl1.downlink.nrPSS import nrPSS
from pynrl1.downlink.nrPSSIndices import nrPSSIndices

def run_nr_pss(ncellid, eng):
    ref_data = eng.nrPSS(matlab.double(ncellid))
    ref_data = np.array(list(itertools.chain(*ref_data)))

    ref_ind = eng.nrPSSIndices()
    ref_ind = np.array(list(itertools.chain(*ref_ind)))
    ref_ind = np.array(ref_ind - 1)

    data = nrPSS(ncellid)
    data = [(x*2)-1 for x in data]

    indices = nrPSSIndices()

    assert np.all(ref_data == data)
    assert np.all(indices == ref_ind)

@pytest.mark.parametrize("ncellid", [0, 500, 1007])
def test_nr_pss(ncellid):
    eng = matlab.engine.connect_matlab()

    try:
        run_nr_pss(ncellid, eng)
    finally:
        eng.quit()



# 38.211

import numpy as np

from pynrl1.util.nrPRBS import nrPRBS
from pynrl1.util.nrSymbolModulate import nrSymbolModulate
from pynrl1.util.nrPDSCHConfig import nrPDSCHConfig
from pynrl1.util.nrCarrierConfig import nrCarrierConfig

def nrPDSCHDMRS(cfg: nrPDSCHConfig, carrier: nrCarrierConfig):
    if cfg.dmrs_conf_type == 1:
        n_dmrs_per_re = 6
    else:
        n_dmrs_per_re = 4
    n_dmrs_bits_re = 2*n_dmrs_per_re

    dmrs_begin = n_dmrs_bits_re * min(cfg.PRB_set)
    dmrs_end = n_dmrs_bits_re * (max(cfg.PRB_set)+1)
    dmrs_size = n_dmrs_bits_re * cfg.n_size_bwp

    n_scid = 0

    occupied_syms = pdschdmrs_occ_symbols(cfg.dmrs_typeA_pos, cfg.symbol_allocation[1], cfg.dmrs_additional_pos)

    # Start generation for every symbol
    dmrs_syms = np.array([])
    for n_symb in occupied_syms:
        cinit_dmrs = nr_pdschdmrs_cinit(carrier.symbols_per_slot, carrier.n_slot, n_symb, cfg.dmrs_NIDNSCID, n_scid)
        dmrs_prbs = nrPRBS(cinit_dmrs, dmrs_size)

        # Cut PRBS sequency
        dmrs_prbs = dmrs_prbs[dmrs_begin:dmrs_end]
        dmrs_syms = np.append(dmrs_syms, nrSymbolModulate(dmrs_prbs, "QPSK"))

    return dmrs_syms

# LUT for DMRS occupied symbols positions
def pdschdmrs_occ_symbols(typeA_pos, sym_alloc, add_pos):
    l1 = 11

    occupied_syms = np.array([], dtype=int)
    occupied_syms = np.append(occupied_syms, typeA_pos)

    if sym_alloc in [8, 9]:
        occupied_syms = np.append(occupied_syms, 7)
    elif sym_alloc in [10, 11]:
        if add_pos == 1:
            occupied_syms = np.append(occupied_syms, 9)
        elif add_pos == 2 or add_pos == 3:
            occupied_syms = np.append(occupied_syms, [6, 9])
    elif sym_alloc == 12:
        if add_pos == 1:
            occupied_syms = np.append(occupied_syms, 11)
        elif add_pos == 2:
            occupied_syms = np.append(occupied_syms, [7, 11])
        elif add_pos == 3:
            occupied_syms = np.append(occupied_syms, [5, 8, 11])
    elif sym_alloc in [13, 14]:
        if add_pos == 1:
            occupied_syms = np.append(occupied_syms, l1)
        elif add_pos == 2:
            occupied_syms = np.append(occupied_syms, [7, 11])
        elif add_pos == 3:
            occupied_syms = np.append(occupied_syms, [5, 8, 11])
    return occupied_syms

def nr_pdschdmrs_cinit(sps, n_slot, n_symb, NIDSCID, n_scid):
    return 2**17 * (sps * n_slot + n_symb + 1) * (2*NIDSCID + 1) + 2*NIDSCID + n_scid


import matlab.engine
import itertools
import numpy as np
import pytest

from pynrl1.downlink.nrPBCH import nrPBCH
from pynrl1.downlink.nrPBCHIndices import nrPBCHIndices

def run_nr_pbch(ncellid, databits, v, eng):
    ref_data = eng.nrPBCH(databits, matlab.double(ncellid), matlab.double(v), nargout=1)
    ref_indices = eng.nrPBCHIndices(ncellid);

    ref_data = np.array(list(itertools.chain(*ref_data)))

    ref_indices = np.array(list(itertools.chain(*ref_indices)))
    ref_indices = [x-1 for x in ref_indices]

    indices = nrPBCHIndices(ncellid)
    data = nrPBCH( ncellid, v, np.array(list(itertools.chain(*databits)), dtype=(int)) )

    ref_data = np.around(ref_data, 4)
    data = np.around(data, 4)

    assert (ref_data == data).all()
    assert (ref_indices == indices).all()

@pytest.mark.parametrize("ncellid", [0, 500, 1007])
@pytest.mark.parametrize("v", list(range(8)))
def test_nr_pbch(ncellid, v):
    eng = matlab.engine.connect_matlab()

    try:
        run_nr_pbch(ncellid, eng.randi(matlab.double([0, 1]),864,1), v, eng)
    finally:
        eng.quit()



import numpy as np

def nrPBCHDMRSIndices(ncellid):
    assert ncellid >= 0 and ncellid <= 1007

    # Indices with window for SSS
    K = 240
    ind  = list(range(  K,        2*K, 4)) #l = 1, k =   0 ... 236
    ind += list(range(2*K,     2*K+48, 4)) #l = 2, k =   0 ...  44
    ind += list(range(2*K+192,    3*K, 4)) #l = 2, k = 192 ... 236
    ind += list(range(3*K,        4*K, 4)) #l = 3, k =   0 ... 236

    # DMRS indices shift by "v"
    ind = np.array(ind)+(ncellid % 4)

    return ind




class nrNumerologyConfig:
        def __init__(self):
            self._subcarrier_spacing = 15
            self._cyclic_prefix = "normal"

        @property
        def subcarrier_spacing(self):
            return self._subcarrier_spacing

        @subcarrier_spacing.setter
        def subcarrier_spacing(self, spacing):
            assert spacing in [15, 30, 60, 120, 240], "The value must be [15, 30, 60, 120, 240]"
            self._subcarrier_spacing = spacing

        @property
        def cyclic_prefix(self):
            return self._cyclic_prefix

        @cyclic_prefix.setter
        def cyclic_prefix(self, pref):
            assert pref in ["normal", "extended"], "The value must be [15, 30, 60, 120, 240]"
            self._cyclic_prefix = pref


import numpy as np
from pynrl1.util.nrPRBS import nrPRBS
from pynrl1.util.nrSymbolModulate import nrSymbolModulate

def nrPBCHDMRS(ncellid, issb):
    assert ncellid >= 0 and ncellid <= 1007
    assert issb >= 0 and issb <= 7

    # Get DMRS sequence
    cinit = nr_pbch_dmrs_cinit(issb, ncellid)
    prbs = nrPRBS(cinit, 2*144)

    # Convert sequence to symbols and merge with indices
    return nrSymbolModulate(prbs, 'qpsk')


def nr_pbch_dmrs_cinit(issb, ncellid):
    return 2**11 * (issb + 1) * (ncellid//4 + 1) + 2**6 * (issb + 1) + (ncellid % 4)



# 38.211 7.4.3 SS/PBCH block

import numpy as np

def nrPBCHIndices(ncellid):
    assert ncellid <= 1007 and ncellid >= 0

    # Indices with windows for SSB
    K = 240
    ind  = list(range(  K,        2*K, 4)) #l = 1, k =   0 ... 236
    ind += list(range(2*K,     2*K+48, 4)) #l = 2, k =   0 ...  44
    ind += list(range(2*K+192,    3*K, 4)) #l = 2, k = 192 ... 236
    ind += list(range(3*K,        4*K, 4)) #l = 3, k =   0 ... 236

    # 3 indices for PBCH within groups of 4, depending on DMRS shift
    v_indices = ncellid % 4
    if v_indices == 0:
        dmrs_ind = [1, 2, 3]
    elif v_indices == 1:
        dmrs_ind = [0, 2, 3]
    elif v_indices == 2:
        dmrs_ind = [0, 1, 3]
    elif v_indices == 3:
        dmrs_ind = [0, 1, 2]

    # Compile DMRS+PBCH indices
    res_ind = [j for x in ind for j in [x+dmrs_ind[0], x+dmrs_ind[1], x+dmrs_ind[2]]]

    return np.array(res_ind)



import matlab.engine
import itertools
import numpy as np
import pytest

from pynrl1.util.nrPRBS import nrPRBS

def run_nr_prbs(cinit, size, eng):
    ref_data = eng.nrPRBS(matlab.double(cinit), matlab.double(size))
    ref_data = list(itertools.chain(*ref_data))
    ref_data = np.array([x*1 for x in ref_data])

    data = nrPRBS(cinit, size)

    assert (ref_data == data).all()

@pytest.mark.parametrize("cinit", [0, 100, 1245345, 534667868])
@pytest.mark.parametrize("size", [1024, 8192])
def test_nr_prbs(cinit, size):
    eng = matlab.engine.connect_matlab()

    try:
        run_nr_prbs(cinit, size, eng)
    finally:
        eng.quit()



import matlab.engine
import os
import itertools
import numpy as np
import pytest

from pynrl1.downlink.nrPDSCHDMRS import nrPDSCHDMRS
from pynrl1.downlink.nrPDSCHDMRSIndices import nrPDSCHDMRSIndices
from pynrl1.util.nrPDSCHConfig import nrPDSCHConfig
from pynrl1.util.nrCarrierConfig import nrCarrierConfig

def run_nr_pdschdmrs(cfg, eng):
    carrier = nrCarrierConfig();
    carrier.subcarrier_spacing = 120;
    carrier.cyclic_prefix = 'normal';
    carrier.n_size_grid = 132;
    carrier.n_start_grid = 0;

    pdsch_cfg = nrPDSCHConfig();
    pdsch_cfg.n_size_bwp = cfg['n_size_bwp']
    pdsch_cfg.n_start_bwp = cfg['n_start_bwp']
    pdsch_cfg.mapping_type = cfg['MappingType']
    pdsch_cfg.dmrs_typeA_pos = cfg['DMRSTypeAPosition']
    pdsch_cfg.dmrs_len = cfg['DMRSLength']
    pdsch_cfg.dmrs_additional_pos = cfg['DMRSAdditionalPosition']
    pdsch_cfg.dmrs_conf_type = cfg['DMRSConfigurationType']
    pdsch_cfg.dmrs_NIDNSCID = cfg['NIDNSCID']
    pdsch_cfg.dmrs_NSCID = cfg['NSCID']
    pdsch_cfg.PRB_set = cfg['PRBSet']
    pdsch_cfg.symbol_allocation = cfg['SymbolAllocation']

    pdschdmrs_syms = nrPDSCHDMRS(pdsch_cfg, carrier)
    pdschdmrs_indices = nrPDSCHDMRSIndices(pdsch_cfg)

    [pdschdmrs_syms_ref, pdschdmrs_indices_ref] = eng.gen_pdschdmrs(cfg, nargout=2)
    pdschdmrs_syms_ref = np.array(list(itertools.chain(*pdschdmrs_syms_ref)))

    pdschdmrs_indices_ref = np.array(list(itertools.chain(*pdschdmrs_indices_ref)))
    pdschdmrs_indices_ref = pdschdmrs_indices_ref - 1

    pdschdmrs_syms = np.around(pdschdmrs_syms, 4)
    pdschdmrs_syms_ref = np.around(pdschdmrs_syms_ref, 4)

    assert np.array_equal(pdschdmrs_syms, pdschdmrs_syms_ref)
    assert np.array_equal(pdschdmrs_indices, pdschdmrs_indices_ref)


@pytest.mark.parametrize('typeA_pos', [2, 3])
@pytest.mark.parametrize('symb_alloc', [[2, 12]])
@pytest.mark.parametrize('dmrs_add_pos', [0, 1, 2, 3])
@pytest.mark.parametrize('PRBSet', [list(range(0, 132)), list(range(60, 132)), list(range(30, 60))])
@pytest.mark.parametrize('dmrs_cfg_type', [1, 2])
def test_nr_pdschdmrs(symb_alloc, dmrs_add_pos, typeA_pos, PRBSet, dmrs_cfg_type):
    eng = matlab.engine.connect_matlab()
    eng.cd(os.path.dirname(__file__))

    cfg = {}
    cfg['n_size_bwp'] = 132
    cfg['n_start_bwp'] = 0
    cfg['MappingType'] = "A"
    cfg['DMRSTypeAPosition'] = typeA_pos
    cfg['DMRSLength'] = 1
    cfg['DMRSAdditionalPosition'] = dmrs_add_pos
    cfg['PRBSet'] = PRBSet
    cfg['SymbolAllocation'] = symb_alloc
    cfg['DMRSConfigurationType'] = dmrs_cfg_type
    cfg['NIDNSCID'] = 1
    cfg['NSCID'] = 0

    try:
        run_nr_pdschdmrs(cfg, eng)
    finally:
        eng.quit()

if __name__ == '__main__':
    test_nr_pdschdmrs([2, 12], 1, 2, list(range(2, 130)), 2)



from pynrl1.util.nrNumerologyConfig import nrNumerologyConfig

class nrCarrierConfig(nrNumerologyConfig):
        def __init__(self):
            self._n_cell_id = 1
            self._n_size_grid = 52
            self._n_start_grid = 0
            self._n_slot = 0
            self._n_frame = 0

            self.__symbols_per_slot = 0;
            self.__slots_per_subframe = 0;
            self.__slots_per_frame = 0;

        @property
        def symbols_per_slot(self):
            if self.cyclic_prefix == "normal":
                return 14
            else:
                return 12

        @property
        def slots_per_subframe(self):
            return self.subcarrier_spacing//15

        @property
        def slots_per_frame(self):
            return int(10*(self.subcarrier_spacing/15))

        @property
        def n_cell_id(self):
            return self._n_cell_id

        @n_cell_id.setter
        def n_cell_id(self, nid):
            assert nid >= 0 and nid <= 1007, "The value must be in the range 0...1007."
            self._n_cell_id = nid

        @property
        def n_size_grid(self):
            return self._n_size_grid

        @n_size_grid.setter
        def n_size_grid(self, size):
            assert size > 0 and size <= 275, "The value must be in the range 1...275."
            self._n_size_grid = size

        @property
        def n_start_grid(self):
            return self._n_start_grid

        @n_start_grid.setter
        def n_start_grid(self, start):
            assert start >= 0 and start <= 2199, "The value must be in the range 0...2999."
            self._n_start_grid = start

        @property
        def n_slot(self):
            return self._n_slot

        @n_slot.setter
        def n_slot(self, num):
            assert num >= 0, "The value must be scalar, nonnegative."
            self._n_slot = num

        @property
        def n_frame(self):
            return self._n_frame

        @n_frame.setter
        def n_frame(self, num):
            assert num >= 0, "The value must be scalar, nonnegative."
            self._n_frame = num


import matlab.engine
import itertools
import numpy as np
import pytest

from pynrl1.downlink.nrPBCHDMRS import nrPBCHDMRS
from pynrl1.downlink.nrPBCHDMRSIndices import nrPBCHDMRSIndices

def run_nr_pbch_dmrs(ncellid, issb, eng):
    ref_data = eng.nrPBCHDMRS(matlab.double(ncellid), matlab.double(issb))
    ref_data = np.array(list(itertools.chain(*ref_data)))

    ref_ind = eng.nrPBCHDMRSIndices(matlab.double(ncellid))
    ref_ind = np.array(list(itertools.chain(*ref_ind)))
    ref_ind = np.array(ref_ind - 1)

    data = nrPBCHDMRS(ncellid, issb)
    indices = nrPBCHDMRSIndices(ncellid)

    ref_data = np.around(ref_data, 4)
    data = np.around(data, 4)

    assert np.all(ref_data == data)
    assert np.all(ref_ind == indices)

@pytest.mark.parametrize("ncellid", [0, 500, 1007])
@pytest.mark.parametrize("issb", list(range(8)))
def test_nr_pbch_dmrs(ncellid, issb):
    eng = matlab.engine.connect_matlab()

    try:
        run_nr_pbch_dmrs(ncellid, issb, eng)
    finally:
        eng.quit()




# 38.211

import numpy as np

from pynrl1.downlink.nrPDSCHDMRS import pdschdmrs_occ_symbols
from pynrl1.util.nrPDSCHConfig import nrPDSCHConfig

def nrPDSCHDMRSIndices(cfg: nrPDSCHConfig):
    frame_begin = cfg.n_rb_size * min(cfg.PRB_set)
    frame_end = cfg.n_rb_size * (max(cfg.PRB_set)+1)
    frame_size = cfg.n_rb_size * cfg.n_size_bwp

    # Single frame DMRS positions
    dmrs_full_range = np.array(list(range(frame_begin, frame_end)))

    # Calculate DMRS positions in every occupied symbol. Frequency position
    if cfg.dmrs_conf_type == 1:
        # Takes every 2nd symbol
        occupied_res = dmrs_full_range[::2]
    elif cfg.dmrs_conf_type == 2:
        # Takes every 4th couple of symbols
        occupied_res = dmrs_full_range.reshape(-1, 2)[::3].ravel()

    # Calculates occupied symbols numbers. Time positions
    occupied_syms = pdschdmrs_occ_symbols(cfg.dmrs_typeA_pos, cfg.symbol_allocation[1], cfg.dmrs_additional_pos)

    dmrs_indices = np.array([])
    for idx, sym in enumerate(occupied_syms):
        sym_offset = sym*frame_size
        dmrs_indices = np.append(dmrs_indices, (occupied_res + sym_offset))

    return dmrs_indices



import matlab.engine
import itertools
import numpy as np
import pytest

from pynrl1.downlink.nrSSS import nrSSS
from pynrl1.downlink.nrSSSIndices import nrSSSIndices

def run_nr_sss(ncellid, eng):
    ref_data = eng.nrSSS(matlab.double(ncellid))
    ref_data = np.array(list(itertools.chain(*ref_data)))

    ref_ind = eng.nrSSSIndices()
    ref_ind = np.array(list(itertools.chain(*ref_ind)))
    ref_ind = np.array(ref_ind - 1)

    data = nrSSS(ncellid)
    indices = nrSSSIndices()
    data = [(x*2)-1 for x in data]

    assert (ref_data == data).all()
    assert (indices == ref_ind).all()

@pytest.mark.parametrize("ncellid", [0, 600, 1007])
def test_nr_sss(ncellid):
    eng = matlab.engine.connect_matlab()

    try:
        run_nr_sss(ncellid, eng)
    finally:
        eng.quit()





class nrPDSCHConfig:
     def __init__(self):
          self._n_rb_size = 12
          self._n_size_bwp = 0
          self._n_start_bwp = 0
          self._mapping_type = "A"

          self._dmrs_typeA_pos = 2
          self._dmrs_len = 1
          self._dmrs_additional_pos = 1
          self._dmrs_conf_type = 2
          self._dmrs_port_set = 0
          self._num_CDM_groups_without_data = 1
          self._dmrs_NIDNSCID = 1
          self._dmrs_NSCID = 0
          self._PRB_set = list(range(0, self._n_size_bwp))
          self._symbol_allocation = [2, 12]

     @property
     def n_rb_size(self):
         return self._n_rb_size

     @property
     def n_size_bwp(self):
         return self._n_size_bwp

     @n_size_bwp.setter
     def n_size_bwp(self, size):
         assert size > 1 and size < 275, "The value must be in the range 1...275."
         self._n_size_bwp = size

     @property
     def n_start_bwp(self):
         return self._n_start_bwp

     @n_start_bwp.setter
     def n_start_bwp(self, start):
         assert start >= 0 and start < 2473, "The value must be in the range 0...2473."
         self._n_start_bwp = start

     @property
     def mapping_type(self):
         return self._mapping_type

     @mapping_type.setter
     def mapping_type(self, map_type):
         assert map_type in ["A", "B"], "The value must be in A or B"
         self._mapping_type = map_type

     @property
     def dmrs_typeA_pos(self):
         return self._dmrs_typeA_pos

     @dmrs_typeA_pos.setter
     def dmrs_typeA_pos(self, pos):
         assert pos in [2, 3], "The value must be 2 or 3"
         self._dmrs_typeA_pos = pos

     @property
     def dmrs_len(self):
         return self._dmrs_len

     @dmrs_len.setter
     def dmrs_len(self, length):
         assert length in [1, 2], "The value must be 1 or 2"
         self._dmrs_len = length

     @property
     def dmrs_additional_pos(self):
         return self._dmrs_additional_pos

     @dmrs_additional_pos.setter
     def dmrs_additional_pos(self, pos):
         assert pos in list(range(0,4)), "The value must be 0..3"
         self._dmrs_additional_pos = pos

     @property
     def dmrs_conf_type(self):
         return self._dmrs_conf_type

     @dmrs_conf_type.setter
     def dmrs_conf_type(self, conf_type):
         assert conf_type in [1, 2], "The value must be 1 or 2"
         self._dmrs_conf_type = conf_type

     @property
     def dmrs_port_set(self):
         assert False, "Not implemented!"
         return self._dmrs_port_set

     @dmrs_port_set.setter
     def dmrs_port_set(self, port_set):
         assert False, "Not implemented!"
         assert port_set in list(range(0, 11)), "The value must be 0..1"
         self._dmrs_port_set = port_set

     @property
     def num_CDM_groups_without_data(self):
         assert False, "Not implemented!"
         return self._num_CDM_groups_without_data

     @num_CDM_groups_without_data.setter
     def num_CDM_groups_without_data(self, num):
         assert False, "Not implemented!"
         assert num in list(range(0, 11)), "The value must be 0..1"
         self._num_CDM_groups_without_data = num

     @property
     def dmrs_NIDNSCID(self):
         return self._dmrs_NIDNSCID

     @dmrs_NIDNSCID.setter
     def dmrs_NIDNSCID(self, nid):
         assert nid > 0 and nid < 65535, "The value must be 0..65535"
         self._dmrs_NIDNSCID = nid

     @property
     def dmrs_NSCID(self):
         return self._dmrs_NIDNSCID

     @dmrs_NSCID.setter
     def dmrs_NSCID(self, nscid):
         assert nscid in [0, 1], "The value must be 0 or 1"
         self._dmrs_NSCID = nscid

     @property
     def PRB_set(self):
         return self._PRB_set

     @PRB_set.setter
     def PRB_set(self, prbset):
         assert min(prbset) >= 0 and max(prbset) < self._n_size_bwp, "The value must be 0..n_size_bwp"
         self._PRB_set = prbset

     @property
     def symbol_allocation(self):
         return self._symbol_allocation

     @symbol_allocation.setter
     def symbol_allocation(self, sym_list):
         # assert prbset > 0 and prbset < self._n_size_bwp, "The value must be 0..n_size_bwp"
         self._symbol_allocation = sym_list



# 38.211 5.2.1 Pseudo-random sequence generation

import numpy as np

def nrPRBS(cinit, n):

    assert n > 0

    # 9 = [1 0 0 1], 15 = [1 1 1 1] (both arranged msb->lsb)
    poly = [9, 15]

    # Register initialization
    reg = [1, cinit]

    # Pre-computed output masks for 1600 shift initialization
    masks = [35263098, 10031374]

    # Memory to store shift register value across the sequence
    seqpair = np.zeros((n, 2), dtype=(int))
    ts = np.zeros((n, 2), dtype=(int))

    feedback = [0, 0]

    # Parity table
    Parity15 = [x*(2**(31-1))
                for x in [0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0]]

    # Run the shift registers
    for i in range(n):
        # Store the masked register values
        seqpair[i][0] = reg[0] & masks[0]
        seqpair[i][1] = reg[1] & masks[1]
        # print(seqpair)

        # Top bit of 32 bit shift register
        and_regpoly = np.bitwise_and(reg, poly)
        # print(and_regpoly)

        feedback[0] = Parity15[int(and_regpoly[0])]
        feedback[1] = Parity15[int(and_regpoly[1])]

        # Shift registers down to the right
        reg[0] = reg[0] >> 1
        reg[1] = reg[1] >> 1

        # Then shift in top bit
        reg[0] = np.bitwise_xor(reg[0], feedback[0])
        reg[1] = np.bitwise_xor(reg[1], feedback[1])

    # Parity table for output masking
    Parity = np.array([
        0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
        1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
        1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
        0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
        1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
        0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
        0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
        1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0
    ])

    # Output masking to implement the shift
    pseq = np.zeros((n, 2), dtype=(int))
    # Shift down 16 and combine
    for i in range(n):
        ts[i][0] = np.bitwise_xor(seqpair[i][0], seqpair[i][0] >> 16)
        ts[i][1] = np.bitwise_xor(seqpair[i][1], seqpair[i][1] >> 16)

        # Shift down 8 and combine
        ts[i][0] = np.bitwise_xor(ts[i][0], ts[i][0] >> 8)
        ts[i][1] = np.bitwise_xor(ts[i][1], ts[i][1] >> 8)

        # Mask lower 8 bits
        ts[i][0] = np.bitwise_and(ts[i][0], 255)
        ts[i][1] = np.bitwise_and(ts[i][1], 255)

        # Look up parity value for these 8 bits
        pseq[i][0] = Parity[ts[i][0]]
        pseq[i][1] = Parity[ts[i][1]]

    # Combine (xor) together the two PN sequences
    return np.array([np.bitwise_xor(a, b) for (a, b) in pseq])



import matlab.engine
import numpy as np
import pytest
import itertools

from pynrl1.util.nrSymbolModulate import nrSymbolModulate

def get_bin_symbols(bps, symbols):
    databits = np.array(list(range(symbols)))
    databits = np.array([(x & (2**np.arange(bps)) != 0).astype(int) for x in databits])
    return np.concatenate(databits, axis=0)

def run_nr_mapper(eng, modtype):
    if modtype == 'bpsk':
        databits = get_bin_symbols(1, 2)
        ref_data = eng.nrSymbolModulate(eng.transpose(eng.logical(databits)), "BPSK")
        data = nrSymbolModulate(databits, modtype)

    elif modtype == 'bpsk_pi2':
        databits = get_bin_symbols(1, 2)
        ref_data = eng.nrSymbolModulate(eng.transpose(eng.logical(databits)), "pi/2-BPSK")
        data = nrSymbolModulate(databits, modtype)

    elif modtype == 'qpsk':
        databits = get_bin_symbols(2, 4)
        ref_data = eng.nrSymbolModulate(eng.transpose(eng.logical(databits)), "QPSK")
        data = nrSymbolModulate(databits, modtype)

    elif modtype == 'qam16':
        databits = get_bin_symbols(4, 16)
        ref_data = eng.nrSymbolModulate(eng.transpose(eng.logical(databits)), "16QAM")
        data = nrSymbolModulate(databits, modtype)

    elif modtype == 'qam64':
        databits = get_bin_symbols(6, 64)
        ref_data = eng.nrSymbolModulate(eng.transpose(eng.logical(databits)), "64QAM")
        data = nrSymbolModulate(databits, modtype)

    elif modtype == 'qam256':
        databits = get_bin_symbols(8, 256)
        ref_data = eng.nrSymbolModulate(eng.transpose(eng.logical(databits)), "256QAM")
        data = nrSymbolModulate(databits, modtype)

    ref_data = np.around(np.array(list(itertools.chain(*ref_data))), 4)
    data = np.around(data, 4)

    assert (ref_data == data).all()

@pytest.mark.parametrize("modtype", ['bpsk', 'bpsk_pi2', 'qpsk', 'qam16', 'qam64', 'qam256'])
def test_nr_pbch(modtype):
    eng = matlab.engine.connect_matlab()

    try:
        run_nr_mapper(eng, modtype)
    finally:
        eng.quit()
